freeze;

/***
 *  Galois Pseudo Primality Test
 *
 *  Distributed under the terms of the GNU Lesser General Public License (L-GPL)
 *                  http://www.gnu.org/licenses/
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published by
 *  the Free Software Foundation; either version 2.1 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 *  Copyright 2012  J.-M. Couveignes, T. Ezome and R. Lercier.
 */

 /***
 *
 * This MAGMA package is an implementation of the Galois pseudo-primality test
 * (see [CEL2012]).
 *
 * The main function is IsGaloisProbablePrime(n : Bases := lambda), that
 * tests if n is a pseudo prime up to probability 4^(-lambda) for some postive
 * integer lambda. 
 *
 * Bibliography :
 * --------------
 *
 * [CEL2012] "A faster pseudo-primality test", J.-M. Couveignes, T. Ezome and
 * R. Lercier, Rendiconti del Circolo  Matematico di Palermo Journal, 18 pages,
 * May 2012, Springer. Available at http://dx.doi.org/10.1007/s12215-012-0088-0.
 * 
 * Example :
 * ---------
 *
 * $ magma
 * > Attach("galoistest.m");
 * > SetVerbose("GaloisTest", 1);
 * > n := 2^4096 - 10895177;
 * > lambda := 1024;
 * > IsGaloisProbablePrime(n : Bases := lambda div 2);
 *
 ********************************************************************************/

 /***
 * Exports :
 *
 * declare verbose GaloisTest, 1;
 *
 * intrinsic IsGaloisProbablePrime(n::RngIntElt : Bases := 20) -> BoolElt
 * 
 ********************************************************************/
 
 /***
 * Changes.
 *
 * 2012-05-28, v1.1 :  J.-M. Couveignes, T. Ezome and R. Lercier,
 *             Reformat the code like a MAGMA package file.
 *
 * 2012-04-06, v1.0 :  J.-M. Couveignes, T. Ezome and R. Lercier,
 *	       Initial version.
 *
 ********************************************************************/

declare verbose GaloisTest, 1;

// Cost estimates
BaseCOST    := 1/(3*10^10);

EXPMR := 2.6;
function CostMR(b)
    return BaseCOST * b^(EXPMR);
end function;

CSG_1 := 1  * BaseCOST;
CSG_2 := 10 * BaseCOST;
EXPPOL := 2.4;
function CostSG(b, dcyc, dkum)
    if dcyc eq 1 then
	return CSG_1 * b^(EXPMR) * dcyc * dkum;
    end if;
    return CSG_2 * b^(EXPPOL) * dcyc * dkum;
end function;

CXN_1 := 19 * BaseCOST;
CXN_2 := 36 * BaseCOST;
EXPLIN := 1.2;
function CostXN(b, dcyc, dkum)
    if dcyc eq 1 then
	return CXN_1 * b^(EXPPOL)* (dcyc * dkum)^(EXPLIN);
    end if;
    return CXN_2  * b^(EXPPOL)* (dcyc * dkum)^(EXPLIN);
end function;

function CostF(b, dcyc)
    if dcyc eq 1 then
	return 0;
    end if;
    if dcyc eq 2 then
	return Ceiling(Log(2, b))*CostMR(b);
    end if;
    return Ceiling(Log(2, dcyc)) * (dcyc div 2) * CostXN(b, dcyc, 1);
end function;

function CostGaloisTest(b, r, dcyc, dkum)
    return r*CostMR(b) +
	CostF(b, dcyc) +
	dcyc*CostXN(b, dcyc, 1) +
	CostSG(b, dcyc, dkum) +
	CostXN(b, dcyc, dkum);
end function;

// Usual Miller Rabin test
function MyMillerRabinTest(R : Bases := 1)

    if #R eq 2 or #R eq 3 then return true; end if;
    
    if IsEven(#R) then return false; end if;
    
    s := 0; d := #R-1; while d mod 2 eq 0 do s +:= 1; d := d div 2; end while;

    for k := 1 to Bases do
	a := R!(2+Random(#R-4));
	x := (R!a)^d;
	if x eq R!1 or x eq -R!1 then continue k; end if;
	for r := 1 to s-1 do
	    x := x^2;
	    if x eq R!1 then return false; end if;
	    if x eq -R!1 then continue k; end if;
	end for;
	return false;
    end for;
    
    return true;

end function;

// Testing the divisibility up to B
function MyTrialDivision(N, B)

    LST := []; p := 2; _N := N;
    while p lt B and _N gt 1 do 

	pow := 0; while (_N mod p) eq 0 do _N div:= p; pow +:= 1; end while;
	if pow gt 0 then Append(~LST, <p, pow>); end if;

	p := NextPrime(p);
    end while;
    
    return LST;
end function;

// ``Irreducible'' polynomial of degree d modulo n
function ExtensionRing(R, d : _a := R!0)

    n := #R;
    X := PolynomialRing(R).1;

    /* Degree 1 : easy case */
    if d eq 1 then
	ER := ExtensionField<R, X | X-1 : Optimize := false, Check := false>;
	return ER, [ER.1];
    end if;

    /* Degree 2 : Look for an non residue quadratic */
    if d eq 2 then
	a := _a; repeat  a -:= 1; until JacobiSymbol(IntegerRing()!a, n) eq -1;
	ER := ExtensionField<R, X | X^2-a : Optimize := false, Check := false>;
	return ER, ER.1^n;
    end if;

    /* Does d divides n-1 ? */
    if (#R-1) mod d eq 0 then

	a := _a; repeat
	    a -:= 1; b := a^((n-1) div d);
	until &and[(b^(d div dg) - 1) ne 0 : dg in PrimeDivisors(d)]  eq true;

	ER := ExtensionField<R, X | X^d-a : Optimize := false, Check := false>;

	return ER, ER.1^n;
    end if;
    
    /* Ben-Or's algorithm */
    a := _a; repeat  a -:= 1;
	ER := ExtensionField<R, X | X^d+X-a : Optimize := false, Check := false>;
	A := ER.1; isirreduc := true;
	for i := 1 to (d div 2) do
	    A := A^n; if i eq 1 then An := A; end if;
	    PA := Parent(X)!Eltseq(A);
	    if GCD(PA-X, X^d+X-a) ne 1 then
		isirreduc := false; break;
	    end if;
	end for;	    
    until isirreduc;
    return ER, An;

end function;

// Parameters (degree extensions)
function GaloisParameters(n, lambda : CB   := 1 - 1/2000)

    b :=  Ceiling(Log(2, n+1));
    A := 2.0*(1+lambda/(b-1))/CB;
    
    /* Miller Rabin test */
    r := (1+lambda) div 2; dcyc := 1; dkum := 1; TT := r*CostMR(b);
    vprintf GaloisTest, 1 : "%o Miller Rabin tests would cost at least %.2o\n", r, TT;

    dcB := Ceiling((9+Log(b))^(1.5*Max(1, Log(Log(Log(68*Sqrt(Log(2, n))))))));//Ceiling(Log(2, Log(2, n)/2));
    vprintf GaloisTest, 1 : "Now, let us consider increasing dcyc from 1 to %o.\n", dcB;

    for dc := 1 to dcB do
	MTD := MyTrialDivision(n^dc-1, Ceiling(2*Sqrt(A*lambda)/dc));
	if #MTD eq 0 then continue; end if;
	
	Q := &*[c[1]^c[2] : c in MTD];
	DQ := Divisors(Q);
	for dk in DQ do

	    d := dc*dk;

	    if d le Floor(A) then continue; end if;
	    
	    rc := Ceiling(lambda*A/(d-A));
	    
	    tt := CostGaloisTest(b, rc, dc, dk);
	    if tt lt TT then
		vprintf GaloisTest, 1 : "   For r = %o, dcyc = %o, dkum = %o, estimated cost is %.2o [A = %.2o]\n", rc, dc, dk, tt, A;
		dcyc := dc; dkum := dk; r := rc; TT := tt;
	    end if;
	end for;
    end for;

    if dcyc*dkum eq 1 then
	vprintf GaloisTest, 1 : "   Miller Rabin tests are the fastest way to check that n is a pseudo-prime.\n";
    end if;
    
    return r, dcyc, dkum;

end function;

/* B := 8000; PRODPRIMES := &*[ p : p in [2..B] |IsPrime(p) ]; */
PRODPRIMES := 
1362562009103927516181790697908402630750268671988097668597117777450735527854639\
5635761296188217989920144728330170482205594877873323448872822307209040478524718\
9174175334620438629491816815177289021995133227065829836266062818604151245177205\
6484590387955087269949474982621146017318119491514546991869648271932400619948094\
4560040438630226359682911205635904456756209010824071031556329868067046225323608\
4099682582255811267276264276934547103544591529071384627017769994656999032884526\
3378446890842244919134227178993535032097598933440284513533156850116646864071210\
8873030645723994540447982153561221837685000169755274528111208290553535088735846\
1610677217943115917682987142379593583720878144910592984774222885862344696511274\
4786500825318410061684214243018248670821856737745340051235323187388447395528370\
6415147966900153280021865433461764187864281135549823777477504591540988164790224\
9507474994881591625442535984126583412783354604745666944258992658884444492283092\
7760126283633937631418702992007132164919929784148317727124333587746382714421216\
5436121971501376225852796803914550048749666377069665164952007562269434527756606\
4296185920137209619409088017864437332722936454755509481914116442678161585686354\
2701027014969283148506566268029740493028162373587793880672289813797933830108657\
0076420538459726622460516005297799475657038280233158595920964579177327295919959\
0304005190064605086810751217898224670859034739058570448320786357477255337458542\
5239528990706699173928105285592228478808933344599927710007262458948996293160335\
7971915031649527547744719164133240170782936899473370920974675286770486576513468\
3880864962150062336135942192254310104032383098926575055180377708904497055480584\
5167358790289673219052375937503614023926370043950397473459412342796452229268349\
6574287217342484231703157756604406932817710235469543397129765130999198985475134\
0703428302565593238405468587055949073206470156136680917465266689120930400056627\
0609888954422544643269996778334022627705080772051686234408177703504523886131388\
2152840578076120739219356867084481023255178653475326980850258121491370085514097\
6166054697702818262890498416452248041376316347344493619041078854500244509871696\
3737426343654343863804518392328900035744402556546554738042170307121476048759772\
0149182779845513825808028180966837081108402035144562941966094039040399680856136\
8472813138380945488624466869258019942315399035516129267839170292364153892884507\
6539952784517286775221562951457542118453128350253051648156518303972918601076931\
8308117199595370234973564108915909282471948783789936775859752617784056873904005\
7383781416792399461939310328319757398591714475288522742061096609725290084550499\
7761776165937149565279201367059514630079398226833914594640925654594099332130992\
8906743500094334501909264473717441959463167286027621504464819527611267985264650\
0838739816954073757300595129322803296859665036924945143853437342177195179400345\
8284225722862311693838382865417712259815218216427261492909937096315326034991252\
1390508637458455523175784014921337367609599038449395175709143215846072591848385\
8859850472342079012068286200235455811918433433565400509381200091655641087209945\
5799342256430716777000794024994267080595230550559599878287525904667925309341545\
7927849916952400396782227142543100709987336707925846659446636000741629600314154\
8200974796660034302861398913156395253288323717614266201048583531684297324503140\
3070650871264359968240370714635457273644027101396300083268726582240179003628517\
080189317969878575736790;

// The main function
intrinsic IsGaloisProbablePrime(n::RngIntElt : Bases := 20) -> BoolElt
{True iff the integer n is a strong probable prime up to probability 4^(-Bases).}

    TT := Cputime();

    lambda := 2*Bases;
    b := Ceiling(Log(2, n+1));

    vprintf GaloisTest, 1 : "\nChecking the primality of a %o-bit integer, up to probability 1/2^%o\n\n", b, lambda;
    
    /* Has n small prime factors  ? */
    t := Cputime();
    vprintf GaloisTest, 1 : "Has n small prime factors : ";
    isprime := GCD(n, PRODPRIMES) eq 1;
    vprintf GaloisTest, 1 : "%o\n", not isprime;
    vprintf GaloisTest, 1 :"Real time: %o s\n\n", Cputime(t);

    if isprime eq false then return false; end if;

    /* Is n an integer power ? */
    t := Cputime();
    vprintf GaloisTest, 1 : "Is n an integer power : ";
    isprime := not IsPower(n);
    vprintf GaloisTest, 1 : "%o\n", not isprime;
    vprintf GaloisTest, 1 :"Real time: %o s\n\n", Cputime(t);

    if isprime eq false then return false; end if;

    /* What are the best extensions ? */
    t := Cputime();
    r, dcyc, dkum := GaloisParameters(n, lambda);
    vprintf GaloisTest, 1 :"Real time: %o s\n\n", Cputime(t);

    /* Miller Rabin tests
    */
    t := Cputime();
    R := IntegerRing(n);
    vprintf GaloisTest, 1 : "Is n a MillerRabin pseudo prime (%o bases, estimated cost = %.2o): ", r, r*CostMR(b);
    isprime := MyMillerRabinTest(R : Bases := r);
    vprintf GaloisTest, 1 : "%o\n", isprime;
    TMR := (1+lambda)*Cputime(t)/2/r;
    vprintf GaloisTest, 1 :"Real time: %o s\n\n", Cputime(t);

    /* Do we have a galois test to perform ? */
    if dcyc*dkum eq 1 or isprime eq false then return isprime; end if;
   
    /*  From now, we ask to Magma to consider Z/nZ as a field, even if this
        not the case, so that we can easily construct algebras over Z/nZ.
    */
    R := FiniteField(n, 1 : Check:= false, Optimize:= false);
 
    if dcyc eq 1 then

	/* Trivial case : first extension is just Z/nZ */
	Rcyc := R;
	SigmaCMat := Matrix([[R!1]]);
	u := Rcyc!2;

    else

	ca := R!0;
	repeat

	    /* Looking for an irreducible pol. of deg. dcyc */
	    t := Cputime();
	    vprintf GaloisTest, 1 : "Constructing an extension Rcyc of deg. %o [estimated cost = %.2o]\n", dcyc, CostF(b, dcyc);
	    Rcyc, An := ExtensionRing(R, dcyc : _a := ca);
	    F := DefiningPolynomial(Rcyc); ca := -Coefficient(F, 0);
	    PR<X> := Parent(F); X := PR.1;
	    vprintf GaloisTest, 1 : "This extension is defined by F = %o\n", F;
	    vprintf GaloisTest, 1 :"Real time: %o s\n\n", Cputime(t);

	    /* Computing the automorphism Sigma_cyc */
	    vprintf GaloisTest, 1 : "Constructing the Galois automorphism sigma_cyc\n";
	    t := Cputime();
	    Ac := Rcyc!1; SigmaC := [Eltseq(Ac)];
	    for i := 1 to dcyc-1 do Ac *:= An; Append(~SigmaC, Eltseq(Ac)); end for;
	    SigmaCMat := Matrix(SigmaC);
	    vprintf GaloisTest, 1 :"Real time: %o s\n\n", Cputime(t);

	    /* Checking that Sigma_cyc is a morphism of algebra */
	    vprintf GaloisTest, 1 : "Do we have sigma_cyc(alpha^i) = alpha^(n*i) for i = %o ... %o : ", dcyc, 2*dcyc-2;
	    t := Cputime();
	    isprime := true; for i := dcyc to 2*dcyc-2 do
		Ac *:= An;
		isprime and:= Rcyc!Eltseq(Vector(dcyc, Eltseq(Rcyc.1^i))*SigmaCMat) eq Ac;
	    end for;
	    vprintf GaloisTest, 1 : "%o\n", isprime;
	    vprintf GaloisTest, 1 :"Real time: %o s\n\n", Cputime(t);

	    if isprime eq false then
		if MyMillerRabinTest(R : Bases := 1) eq false then return false; end if;
		vprintf GaloisTest, 1 : "We met a non invertible element ...\n";
		vprintf GaloisTest, 1 : "Still, an additional Miller Rabin test is ok. Let us try again ?\n";
		continue;
	    end if;
	    
	    /* Checking that Sigma_cyc fixes exactly Z/nZ */
	    t := Cputime();
	    vprintf GaloisTest, 1 : "Is  Z/nZ the fixed submodule by  sigma_cyc : ";
	    VFixed :=  Eltseq(Basis(Kernel(SigmaCMat - Parent(SigmaCMat)!1))[1]);
	    isprime := VFixed eq [R!1] cat [0 : i in [1..(dcyc-1)]];
	    vprintf GaloisTest, 1 : "%o\n", isprime;
	    vprintf GaloisTest, 1 :"Real time: %o s\n\n", Cputime(t);

	    if isprime eq false then return false; end if;

	    /* Checking that Sigma_cyc is Galois */
	    t := Cputime();
	    vprintf GaloisTest, 1 : "Do we have a Galois ring extension : ";
	    u := Rcyc.1; //+1;
	    nbtries := 0; repeat
		nbtries +:= 1; uc := u;
		isprime := true; iscomposite := false; for i := 1 to dcyc-1 do
		    uc := Rcyc!Eltseq(Vector(dcyc, Eltseq(uc))*SigmaCMat);
		    if uc-u eq 0 then
			isprime := false; break i;
		    else
			isunit, y := IsInvertible(uc - u);
			if not isunit or not y*(uc - u) eq 1 then
			    iscomposite := true; isprime := false; break i;
			end if;
		    end if;
		end for;
		if iscomposite eq true then
		    if MyMillerRabinTest(R : Bases := 1) eq false then return false; end if;
		    vprintf GaloisTest, 1 : "We met a non invertible element ...\n";
		    vprintf GaloisTest, 1 : "Still, an additional Miller Rabin test is ok. Let us try again ?\n";
		    continue;
		end if;
		if isprime eq false then u +:= 1; continue; end if;
		break;
	    until false;
	    vprintf GaloisTest, 1 : "%o (%o tries)\n", isprime, nbtries;
	    vprintf GaloisTest, 1 : "The witness is u = %o\n", u;
	    vprintf GaloisTest, 1 : "Real time: %o s\n\n", Cputime(t);

	    /* sigma^dcyc (x) = x ? */
	    t := Cputime();
	    vprintf GaloisTest, 1 : "Do we have a sigma^dcyc (x) = x : ";
	    if u eq Rcyc.1 then
		uc := Rcyc!Eltseq(Vector(dcyc, Eltseq(uc))*SigmaCMat);
	    else
		uc := Rcyc.1;
		for i := 1 to dcyc do
		    uc := Rcyc!Eltseq(Vector(dcyc, Eltseq(uc))*SigmaCMat);
		end for;
	    end if;
	    isprime := uc eq Rcyc.1;
	    vprintf GaloisTest, 1 : "%o\n", isprime;
	    vprintf GaloisTest, 1 :"Real time: %o s\n\n", Cputime(t);

	    if isprime eq false then return false; end if;
	    
	    break;
	until false;
    end if;

    /* Checking that dkum divides n^dcyc-1 */
    t := Cputime();
    vprintf GaloisTest, 1 : "Does dkum divides n^dcyc-1 : ";
    isprime := (n^dcyc-1) mod dkum eq 0;
    vprintf GaloisTest, 1 : "%o\n", isprime;
    vprintf GaloisTest, 1 :"Real time: %o s\n\n", Cputime(t);
    
    if isprime eq false then return false; end if;

    /* Can we find a pseudo-primitive element in Rcyc */
    t := Cputime(); DK := PrimeDivisors(dkum);
    vprintf GaloisTest, 1 : "Can we find a pseudo-primitive element a  [estimated cost = %.2o]: ", dcyc*CostXN(b, dcyc, 1);
    a := u;
    nbtries := 0; repeat
	nbtries +:= 1;
	zeta := a^((n^dcyc-1) div dkum);
	isprime := true; iscomposite := false; for q in DK  do
	    v := zeta^(dkum div q);
	    if v eq 1 then
		isprime := false; break q;
	    else
		isunit, y := IsInvertible(v-1);
		if not isunit or not y*(v-1) eq 1 then
		    iscomposite := true; isprime := false; break q;
		end if;
	    end if;
	end for;
	if iscomposite eq true then
	    if MyMillerRabinTest(R : Bases := 1) eq false then return false; end if;
	    vprintf GaloisTest, 1 : "%o (%o tries)\n", isprime, nbtries;
	    vprintf GaloisTest, 1 : "We met a non invertible element ...\n";
	    vprintf GaloisTest, 1 : "Still, an additional Miller Rabin test is ok. Let us try again ?\n";
	    continue;
	end if;
	if isprime eq false then a +:= 1; continue; end if;
	vprintf GaloisTest, 1 : "%o (%o tries)\n", isprime, nbtries;

	/* Checking that Sigma_cyc(a) = a^n */
	vprintf GaloisTest, 1 : "Do we have sigma_cyc(a) = a^n : ";
	if dcyc eq 1 then an := a*v^DK[#DK]; else an := a^n; end if;
	isprime :=  Rcyc!Eltseq(Vector(dcyc, Eltseq(a))*SigmaCMat) eq an;
	vprintf GaloisTest, 1 : "%o\n", isprime;
    
	if isprime eq false then return false; end if;

	break; until false;
    vprintf GaloisTest, 1 : "We found a = %o\n", a;
    vprintf GaloisTest, 1 :"Real time: %o s\n\n", Cputime(t);

    /* Constructing the Kummer extension Rkum */
    t := Cputime();
    vprintf GaloisTest, 1 : "Constructing the Kummer extension of degree %o\n", dkum;
    PRcyc<Y> := PolynomialRing(Rcyc); Y := PRcyc.1;
    Rkum<beta> := ExtensionField<Rcyc, Y | Y^dkum-a : Optimize := false, Check := false>;
    vprintf GaloisTest, 1 :"Real time: %o s\n\n", Cputime(t);

    repeat
	
	/* Can we find a random invertible element x in Rkum */
	t := Cputime();
	vprintf GaloisTest, 1 : "Can we find a random unit x in this extension : ";
	repeat x := Random(Rkum); until x ne 0;
	isprime, y := IsInvertible(x);
	if isprime and not y*x eq 1 then isprime := false; end if;
	vprintf GaloisTest, 1 : "%o\n", isprime;
	vprintf GaloisTest, 1 :"Real time: %o s\n\n", Cputime(t);

	if isprime eq false then return false; end if;

	break; until false;

    /* Computing Sigma(x) */
    t := Cputime();
    vprintf GaloisTest, 1 : "Computing sigma(x)    [estimated cost = %.2o]\n", CostSG(b, dcyc, dkum);
    XX := [Rcyc!Eltseq(Vector(dcyc, Eltseq(c))*SigmaCMat) : c in Eltseq(x)];
    SX := &+[XX[i]*a^( ((i-1)*n) div dkum)*beta^(((i-1)*n) mod dkum) : i in [1..#XX] ];
    vprintf GaloisTest, 1 :"Real time: %o s\n\n", Cputime(t);

    /* Computing x^n */
    t := Cputime();
    vprintf GaloisTest, 1 : "Computing x^n         [estimated cost = %.2o]\n", CostXN(b, dcyc, dkum);
    XN := x^n;
    vprintf GaloisTest, 1 :"Real time: %o s\n\n", Cputime(t);
    
    /* The test */
    isprime := SX eq XN;
    
    vprintf GaloisTest, 1 : "Galois primality test : %o\n\n", isprime;
    
    vprintf GaloisTest, 1 : "Total time = %.2o s   [estimated cost = %.2o]\n",
	Cputime(TT), CostGaloisTest(b, r, dcyc, dkum);
    vprintf GaloisTest, 1 : "            (estimated time for a full Miller Rabin test = %.2o s)\n\n", TMR;

    return isprime;

end intrinsic;
